/*
  RabidVWM LiteStep Module - version 1.0
	PixelBlock class and header for Pixel-level Bitmap Manipulation

  Copyright (C) 2002 Joshua Seagoe
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#ifndef TRANSPARENT_COLOR
#define TRANSPARENT_COLOR RGB(255,0,255)
#endif

// this class provides access to bitmaps
// in a way that allows me to change the method later
class PixelBlock {
	HBITMAP hBmp;
	BITMAP Bmp;

	PixelBlock(void);
protected:
public:
	PixelBlock(PixelBlock &src);
	PixelBlock(HBITMAP hbm);
	PixelBlock(DWORD width,DWORD height);
	~PixelBlock();

	PixelBlock &operator=(PixelBlock &source);

	COLORREF *Pixel(int x,int y);
	COLORREF GetPixel(int x,int y) { COLORREF *p=Pixel(x,y); return p?*p:0; }
	bool SetPixel(int x,int y,COLORREF color)
		{ COLORREF *p=Pixel(x,y); return p?(*p=color,true):false; }

	HBITMAP Release(void); // get & remove handle
	HBITMAP Handle(void) // get handle temporarily
		{ return hBmp; }

	COLORREF *GetBits(void) { ASSERTV(hBmp!=NULL); return (COLORREF *)Bmp.bmBits; }
	DWORD GetHeight(void) { ASSERTV(hBmp!=NULL); return Bmp.bmHeight; }
	DWORD GetWidth(void) { ASSERTV(hBmp!=NULL); return Bmp.bmWidth; }
	DWORD GetWidthBytes(void) { ASSERTV(hBmp!=NULL); return Bmp.bmWidthBytes; }
};

// iterate over a PixelBlock
class PixelBlockIterator {
	PixelBlockIterator(void);
	enum {
		RedMask   = (DWORD)-1 - RGB(255,  0,  0),
		GreenMask = (DWORD)-1 - RGB(  0,255,  0),
		BlueMask  = (DWORD)-1 - RGB(  0,  0,255),
		AlphaMask =             RGB(255,255,255)
	};
protected:
	PixelBlock &src;
	COLORREF *ptr;
	DWORD x,y;
	RECT area;
public:
	PixelBlockIterator(PixelBlock &s)
		: src(s) { SetArea(); }
	PixelBlockIterator(PixelBlock &s,RECT &a)
		: src(s), area(a) { ClipArea(); Reset(); }

	void Reset(void) { ptr = src.Pixel(x=area.left,y=area.top); }
	void SetArea(RECT &a) { area = a; ClipArea(); Reset(); }
	void SetArea(void) {
		area.left = area.top = 0;
		area.right = src.GetWidth();
		area.bottom = src.GetHeight();
		Reset();
	}

	void ClipArea(void) {
		RECT a(area);
		RECT b = { 0, 0, src.GetWidth(), src.GetHeight() };
		IntersectRect(&area,&a,&b);
	}

	void operator++(void);

	bool Done(void) { return NULL==ptr; }

	// they're backwards...
	int Red  (void) { ASSERTV(!Done()); return GetBValue(Get()); }
	int Green(void) { ASSERTV(!Done()); return GetGValue(Get()); }
	int Blue (void) { ASSERTV(!Done()); return GetRValue(Get()); }
	int Alpha(void) { ASSERTV(!Done()); return Get()>>24; }
	COLORREF Get(void) { ASSERTV(!Done()); return *ptr; }
	COLORREF Get(int x,int y);
	bool Transparent(void) { ASSERTV(!Done()); return TRANSPARENT_COLOR==Get() || 0==Alpha(); }

	void Set(COLORREF c) { ASSERTV(!Done()); *ptr = c; }
	// remember: Alpha should be premultiplied!
	void Set(int r,int g,int b,int a=0) { ASSERTV(!Done()); *ptr = RGB(b,g,r)+(a<<24); }
	void SetRed  (int r) { ASSERTV(!Done()); *ptr = *ptr&RedMask  |RGB(r,0,0); }
	void SetGreen(int g) { ASSERTV(!Done()); *ptr = *ptr&GreenMask|RGB(0,g,0); }
	void SetBlue (int b) { ASSERTV(!Done()); *ptr = *ptr&BlueMask |RGB(0,0,b); }
	void SetAlpha(int a) { ASSERTV(!Done()); *ptr = *ptr&AlphaMask|(a<<24); }

	void MultiplyAlpha(void) {
		if (Alpha()!=255)
			Set(
				Red()  *Alpha()/255,
				Green()*Alpha()/255,
				Blue() *Alpha()/255,
				Alpha()
			);
	}
};

