/*
  RabidVWM LiteStep Module - version 1.0
	Debug Dump (TRACE is in RabidVWM.cpp)

  Copyright (C) 2002 Joshua Seagoe
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include "RabidVWM.h"

#if !defined(NDEBUG) && defined(TRACE_FILE)

void DumpWindow(const char *level,const s_WinData *pData)
{
	char buffer[256],c[32];;
	if (pData==NULL) {
		TRACE(level,"W[NULL]");
		return;
	}
	GetClassName(pData->hWnd,c,32);
	sprintf(buffer,"W[%2d] = %08X { %5d, %5d, %5d, %5d } %2d %c%c%c%c%c%c%c%c %08X %08X %2d %s",
		(pData-GlobalData.KnownWindows),
		pData->hWnd,
		pData->Area.left,
		pData->Area.top,
		pData->Area.right,
		pData->Area.bottom,
		pData->HomeDesk,
		pData->bNewWindow?'N':'-',
		pData->bTouch==GlobalData.bWindowTouch?'T':'-',
		pData->bTitleChanged?'C':'-',
		pData->bFlash?'F':'-',
		pData->bHidden?'H':'-',
		pData->bFix?'X':'-',
		pData->bSticky?'S':'-',
		pData->bIgnoreTitleChange?'I':'-',
		pData->hIcon,
    pData->Prev,
    pData->PaintNext,
		c); // wndclass
	TRACE(level,buffer);
}

void DumpWindowList(void)
{
	int i;
	s_WinData *pData;
	char buffer[MAX_LINE_LENGTH];

	TRACE(TRACE_DUMP,"");
	sprintf(buffer,"Dumping KnownWindows: count = %d", GlobalData.nKnownWindows);
	TRACE(TRACE_DUMP,buffer);
	TRACE(TRACE_DUMP," [##] = HWND.... { left., top.., right, bottm } hd NTCFHXSI HICON... HWNDPREV nx");

	pData = GlobalData.KnownWindows;
	i = GlobalData.nKnownWindows;
	while (i) {
		DumpWindow(TRACE_DUMP,pData);
		--i; pData++;
	}
}

extern StarWindowMatchList StarLists[];

void DumpStarLists(void)
{
	StarWindowMatchList *pList;
	StarWindowMatchItem *pItem;
	char buffer[MAX_LINE_LENGTH];

	TRACE(TRACE_DUMP,"");
	TRACE(TRACE_DUMP,"Dumping StarLists:");

	pList = StarLists;
	while (pList->Key != NULL) {
		sprintf(buffer,"%-30s S:%08X C:%08X",
			pList->Key,
			pList->Set,
			pList->Clear);
		TRACE(TRACE_DUMP,buffer);
		pItem = pList->pItem;
		while (pItem!=NULL) {
			sprintf(buffer,"   %-30s %d %s",
				pItem->Key,
				pItem->MatchProcFlag,
				pItem->Args?pItem->Args:"NULL");
			TRACE(TRACE_DUMP,buffer);
			pItem = pItem->pNext;
		}
		TRACE(TRACE_DUMP,"  -----");
		pList++;
	}
}

#ifdef VWM_STATS
volatile s_VWMStats VWMStats = { 0 };

void DumpStats(void)
{
	char buffer[MAX_LINE_LENGTH];

	TRACE(TRACE_DUMP,"");
	TRACE(TRACE_DUMP,"Dumping Statistics:");
	sprintf(buffer,"  Time since module start:     %10d",GetTickCount()-VWMStats.InitStart);
	TRACE(TRACE_DUMP,buffer);
	sprintf(buffer,"       in Module:              %10d = %3d.%03d%%",
		VWMStats.InitTime-VWMStats.CreateTime+VWMStats.WndProcTime,
		MulDiv(VWMStats.InitTime-VWMStats.CreateTime+VWMStats.WndProcTime,100,GetTickCount()-VWMStats.InitStart),
		MulDiv(VWMStats.InitTime-VWMStats.CreateTime+VWMStats.WndProcTime,10000,GetTickCount()-VWMStats.InitStart)%1000
		);
	TRACE(TRACE_DUMP,buffer);
	sprintf(buffer,"       in Init:                %10d",VWMStats.InitTime);
	TRACE(TRACE_DUMP,buffer);
	sprintf(buffer,"       in WndProc:             %10d",VWMStats.WndProcTime);
	TRACE(TRACE_DUMP,buffer);
	sprintf(buffer,"       in WM_CREATE:           %10d",VWMStats.CreateTime);
	TRACE(TRACE_DUMP,buffer);
	sprintf(buffer,"       in Refresh:             %10d",VWMStats.RefreshTime);
	TRACE(TRACE_DUMP,buffer);
	sprintf(buffer,"       in Trace:               %10d",VWMStats.TraceTime);
	TRACE(TRACE_DUMP,buffer);
	sprintf(buffer,"       in TransEdgeBlt:        %10d",VWMStats.EdgeBltTime);
	TRACE(TRACE_DUMP,buffer);
	sprintf(buffer,"       in .snapshot:           %10d",VWMStats.SnapShotTime);
	TRACE(TRACE_DUMP,buffer);
	sprintf(buffer,"       in UpdateKnownWindows:  %10d",VWMStats.UpdateTime);
	TRACE(TRACE_DUMP,buffer);
	sprintf(buffer,"       in Repaint:             %10d",VWMStats.PaintTime);
	TRACE(TRACE_DUMP,buffer);
	sprintf(buffer,"       in DoWindowMatches:     %10d",VWMStats.StarListTime);
	TRACE(TRACE_DUMP,buffer);
	sprintf(buffer,"       in DumpStats:           %10d",GetTickCount()-VWMStats.WndProcStart);
	TRACE(TRACE_DUMP,buffer);
	sprintf(buffer,"  Current WndProc depth:       %10d",VWMStats.WndProcLevel);
	TRACE(TRACE_DUMP,buffer);
	TRACE(TRACE_DUMP,"");
	sprintf(buffer,"Bitmaps allocated: %10d",VWMStats.GDIBitmaps);
	TRACE(TRACE_DUMP,buffer);
	sprintf(buffer,"Brushes allocated: %10d",VWMStats.GDIBrushs);
	TRACE(TRACE_DUMP,buffer);
	sprintf(buffer,"    DCs allocated: %10d",VWMStats.GDIDCs);
	TRACE(TRACE_DUMP,buffer);
	sprintf(buffer,"   Pens allocated: %10d",VWMStats.GDIPens);
	TRACE(TRACE_DUMP,buffer);
}
#endif

void DumpBools(void)
{
	char buffer[MAX_LINE_LENGTH];
	rcbooldef *pdef;
	pdef = RCBools;
	TRACE(TRACE_DUMP,"");
	TRACE(TRACE_DUMP,"Dumping bools:");
	while (pdef->key != 0) {
		sprintf(buffer,"%-20s: %s",pdef->key,GlobalData.*pdef->pItem?"true":"false");
		TRACE(TRACE_DUMP,buffer);
		pdef++;
	}
}

void DumpActions(void)
{
	char buffer[MAX_LINE_LENGTH];
	const char *cmd;
	rcactiondef *pdef;
	pdef = RCActions;
	TRACE(TRACE_DUMP,"");
	TRACE(TRACE_DUMP,"Dumping actions:");
	while (pdef->key != 0) {
		if (!(GlobalData.*pdef->pAction && (cmd = (GlobalData.*pdef->pAction)->Command)))
			cmd = "NULL";
		sprintf(buffer,"%-20s: %s",pdef->key,cmd);
		TRACE(TRACE_DUMP,buffer);
		if (GlobalData.*pdef->pAction) {
			sprintf(buffer,"%10d %3d/%3d (%5d,%5d) %1d",
				(GlobalData.*pdef->pAction)->tActivated,
				(GlobalData.*pdef->pAction)->nActivated,
				(GlobalData.*pdef->pAction)->nCalled,
				(GlobalData.*pdef->pAction)->MousePos.x,
				(GlobalData.*pdef->pAction)->MousePos.y,
				(GlobalData.*pdef->pAction)->Hold);
			TRACE(TRACE_DUMP,buffer);
		}
		pdef++;
	}
}

#endif
