/*
  RabidVWM LiteStep Module - version 1.0
	Solid Color Based Painting Functions

  Copyright (C) 2002 Joshua Seagoe
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include "RabidVWM.h"

// Paint using solid colors or bevel

void PaintVWM_Colors(HDC dc);
static void BeginPaintWnd_Colors(HDC dc,PixelBlock &,void *store);
static void PaintWnd_Colors(HDC dc,PixelBlock &,void *store,s_WinData *pData,RECT *rFull,RECT *rWnd);
static void PaintWndTitle_Colors(HDC dc,PixelBlock &,void *store,s_WinData *pData,RECT *rFull,RECT *rTitle);
static void EndPaintWnd_Colors(HDC dc,PixelBlock &,void *store);

static HBRUSH CreateTransBrush(COLORREF c)
{
	if (c==TRANSPARENT_COLOR)
		return (HBRUSH)GetStockObject(NULL_BRUSH);
	return CreateSolidBrush(c);
}

static HPEN CreateTransPen(COLORREF c)
{
	if (c==TRANSPARENT_COLOR)
		return CreatePen(PS_NULL,1,c);
	return CreatePen(PS_SOLID,1,c);
}

const s_PaintWnd Paint_Colors = {
	&BeginPaintWnd_Colors,
	&PaintWnd_Colors,
	&PaintWndTitle_Colors,
	&EndPaintWnd_Colors
};

struct s_PaintWnd_Colors {
	HBRUSH oldBrush;
	HPEN oldPen;

	HBRUSH brBack;
	HBRUSH brTitle;
	HPEN pBackLight,pBackDark,pBackBorder;
	HPEN pTitleLight,pTitleDark,pTitleBorder;
};

void PaintDesk_Colors(HDC dc, s_DeskData *pDesk, RECT *r)
{
	ASSERT(READABLE(r,sizeof(RECT)));
	ASSERT(READABLE(pDesk,sizeof(s_DeskData)));
	ASSERTV(dc!=NULL);

	/* draw the desk */ 
	if (GlobalData.bBackBevel) {
		Rectangle(dc,r->left,r->top,r->right,r->bottom);
		Frame3D(dc,*r,GlobalData.cBack.Light,GlobalData.cBack.Dark,1);
	} else {
		/* don't draw the edges twice */ 
		if (r->right!=GlobalData.VWMSize.cx) r->right++;
		if (r->bottom!=GlobalData.VWMSize.cy) r->bottom++;
		Rectangle(dc,r->left,r->top,r->right,r->bottom);
		/* restore the edges so they all line up */ 
		if (r->right!=GlobalData.VWMSize.cx) r->right--;
		if (r->bottom!=GlobalData.VWMSize.cy) r->bottom--;
	}
}

void PaintSelect_Colors(HDC dc,s_DeskData *pDesk,RECT *r)
{
	HBRUSH brOld,brSelect;

	ASSERT(READABLE(r,sizeof(RECT)));
	ASSERT(READABLE(pDesk,sizeof(s_DeskData)));
	ASSERTV(dc!=NULL);

	brSelect = CreateTransBrush(GlobalData.cSelect.Flat);
	StatGDIAdd(Brush);
	brOld = (HBRUSH)SelectObject(dc,brSelect);

	if (GlobalData.bSelBevel) {
		Rectangle(dc,r->left,r->top,r->right,r->bottom);
		Frame3D(dc,*r,GlobalData.cSelect.Light,GlobalData.cSelect.Dark,1);
	} else {
		HPEN pOld,pSelect;

		pSelect = CreateTransPen(GlobalData.cSelect.Border);
		StatGDIAdd(Pen);
		pOld = (HPEN)SelectObject(dc,pSelect);

		/* don't draw the edges twice */ 
		if (r->right!=GlobalData.VWMSize.cx) r->right++;
		if (r->bottom!=GlobalData.VWMSize.cy) r->bottom++;
		Rectangle(dc,r->left,r->top,r->right,r->bottom);

		SelectObject(dc,pOld);
		DeleteObject(pSelect);
		StatGDIDel(Pen);
	}

	SelectObject(dc,brOld);
	DeleteObject(brSelect);
	StatGDIDel(Brush);
}
	

void PaintVWM_Colors(HDC dc)
{
	HBRUSH brOld;
	HBRUSH brBack;

	brBack = CreateTransBrush(GlobalData.cBack.Flat);
	StatGDIAdd(Brush);
	brOld = (HBRUSH)SelectObject(dc,brBack);

	/* paint the back if a border is specified */ 

	/* this is to paint areas that are on the border, or otherwise are not covered by a desktop */ 
	/* except it breaks VWMNoDrawSelBack, which I seem to have neglected to document            */ 
	/* it's kinda neat tho, so i think i'd like for it to work                                  */ 
	if (!GlobalData.bNoDrawSelBack)
		Rectangle(dc,0,0,GlobalData.VWMSize.cx,GlobalData.VWMSize.cy);
	if (GlobalData.bBackBevel) {
		RECT r;
		SetRect(&r,0,0,GlobalData.VWMSize.cx,GlobalData.VWMSize.cy);
		Frame3D(dc,r,GlobalData.cBack.Dark,GlobalData.cBack.Light,1);
		/* paint the desktop backs */ 
		PaintVWM_Generic(dc,&PaintDesk_Colors);
	} else {
		HPEN pOld;
		HPEN pBorder;

		pBorder = CreateTransPen(GlobalData.cBack.Border);
		StatGDIAdd(Pen);
		pOld = (HPEN)SelectObject(dc,pBorder);

		/* paint the desktop backs */ 
		PaintVWM_Generic(dc,&PaintDesk_Colors);

		SelectObject(dc,pOld);
		DeleteObject(pBorder);
		StatGDIDel(Pen);
	}

	SelectObject(dc,brOld);
	DeleteObject(brBack);
	StatGDIDel(Brush);
}

void BeginPaintWnd_Colors(HDC dc,PixelBlock &pb,void *store)
{
	s_PaintWnd_Colors *pwc;
	/* allocate space for the brushes and pens we'll use */ 

	ASSERTV(store!=NULL);
	ASSERT(PAINTWND_BUFFER_SIZE>=sizeof(s_PaintWnd_Colors));

	pwc = (s_PaintWnd_Colors *)store;

  ASSERT(WRITEABLE(pwc,sizeof(s_PaintWnd_Colors)));

	/* main windows */ 
	pwc->brBack = CreateTransBrush(GlobalData.cWnd.Flat);
	StatGDIAdd(Brush);
	CTRACE(TRACE_WARNING,!pwc->brBack,"BPW_C: Failed to create brBack.");

	pwc->pBackBorder = CreateTransPen(GlobalData.cWnd.Border);
	StatGDIAdd(Pen);
	CTRACE(TRACE_WARNING,!pwc->pBackBorder,"BPW_C: Failed to create pBackBorder.");

	pwc->pBackDark = CreateTransPen(GlobalData.cWnd.Dark);
	StatGDIAdd(Pen);
	CTRACE(TRACE_WARNING,!pwc->pBackDark,"BPW_C: Failed to create pBackDark.");

	pwc->pBackLight = CreateTransPen(GlobalData.cWnd.Light);
	StatGDIAdd(Pen);
	CTRACE(TRACE_WARNING,!pwc->pBackLight,"BPW_C: Failed to create pBackLight.");

	/* titlebars */ 
	if (GlobalData.bTitlebars) {
		pwc->brTitle = CreateTransBrush(GlobalData.cTitle.Flat);
		StatGDIAdd(Brush);
		CTRACE(TRACE_WARNING,!pwc->brTitle,"BPW_C: Failed to create brTitle.");

		pwc->pTitleBorder = CreateTransPen(GlobalData.cTitle.Border);
		StatGDIAdd(Pen);
		CTRACE(TRACE_WARNING,!pwc->pTitleBorder,"BPW_C: Failed to create pTitleBorder.");

		pwc->pTitleDark = CreateTransPen(GlobalData.cTitle.Dark);
		StatGDIAdd(Pen);
		CTRACE(TRACE_WARNING,!pwc->pTitleDark,"BPW_C: Failed to create pTitleDark.");

		pwc->pTitleLight = CreateTransPen(GlobalData.cTitle.Light);
		StatGDIAdd(Pen);
		CTRACE(TRACE_WARNING,!pwc->pTitleLight,"BPW_C: Failed to create pTitleLight.");
	}

	pwc->oldBrush = (HBRUSH)SelectObject(dc,pwc->brBack);
	pwc->oldPen = (HPEN)SelectObject(dc,pwc->pBackBorder);
}

void PaintWnd_Colors(HDC dc,PixelBlock &pb,void *store,s_WinData *pData,RECT *rFull,RECT *rWnd)
{
	s_PaintWnd_Colors *pwc = (s_PaintWnd_Colors *)store;
	HBRUSH brFlash;

	ASSERTV(store!=NULL);
	ASSERTV(pwc!=NULL);
	ASSERTV(pData!=NULL);
	ASSERTV(rFull!=NULL);
	ASSERTV(rWnd!=NULL);

	/* we'll paint the outer bevel using rFull and the inner with rWnd */ 
	/* this way titlebars will automatically be "on" the border        */ 

	DI;

	if ((pData->bFlash && GlobalData.bFlashState) || pData->bFlashOn) {
		brFlash = CreateTransBrush(GlobalData.cFlash.Flat);
		StatGDIAdd(Brush);
		SelectObject(dc,brFlash);
	}

	/* rectangle also draws the border for us */ 
	Rectangle(dc,rFull->left,rFull->top,rFull->right,rFull->bottom);

	if (GlobalData.bBevel) {
		/* outer bevel */ 
		SelectObject(dc,pwc->pBackLight);
		CI MoveToEx(dc,rFull->left,rFull->bottom-1,NULL);
		CI LineTo(dc,rFull->left,rFull->top); /* left edge */ 
		CI LineTo(dc,rFull->right-1,rFull->top); /* top edge */ 
		SelectObject(dc,pwc->pBackDark);
		CI LineTo(dc,rFull->right-1,rFull->bottom-1); /* right edge */ 
		CI LineTo(dc,rFull->left,rFull->bottom-1); /* bottom edge */ 

		CTRACE(TRACE_WARNING,!i,"PW_C: Failed to draw outer bevel.");
		CI0;

		if (GlobalData.DoubleBevelWidth>=0
			&& (rWnd->right-rWnd->left-3)>GlobalData.DoubleBevelWidth*2
			&& (rWnd->bottom-rWnd->top-3)>GlobalData.DoubleBevelWidth*2)
		{
			/* inner bevel */ 
			SelectObject(dc,pwc->pBackDark);
			CI MoveToEx(dc,rWnd->left+1+GlobalData.DoubleBevelWidth,
				rWnd->bottom-2-GlobalData.DoubleBevelWidth,NULL);
			CI LineTo(dc,rWnd->left+1+GlobalData.DoubleBevelWidth,
				rWnd->top+1+GlobalData.DoubleBevelWidth); /* left edge */ 
			CI LineTo(dc,rWnd->right-2-GlobalData.DoubleBevelWidth,
				rWnd->top+1+GlobalData.DoubleBevelWidth); /* top edge  */ 
			SelectObject(dc,pwc->pBackLight);
			CI LineTo(dc,rWnd->right-2-GlobalData.DoubleBevelWidth,
				rWnd->bottom-2-GlobalData.DoubleBevelWidth); /*  right edge */ 
			CI LineTo(dc,rWnd->left+1+GlobalData.DoubleBevelWidth,
				rWnd->bottom-2-GlobalData.DoubleBevelWidth); /* bottom edge */ 

			CTRACE(TRACE_WARNING,!i,"PW_C: Failed to draw inner bevel.");
		}

		/* select the old pen */ 
		SelectObject(dc,pwc->pBackBorder);
	}

	if ((pData->bFlash && GlobalData.bFlashState) || pData->bFlashOn) {
		SelectObject(dc,pwc->brBack);
		DeleteObject(brFlash);
		StatGDIDel(Brush);
	}
}

void PaintWndTitle_Colors(HDC dc,PixelBlock &pb,void *store,s_WinData *pData,RECT *rFull,RECT *rTitle)
{
	s_PaintWnd_Colors *pwc = (s_PaintWnd_Colors *)store;

	ASSERTV(store!=NULL);
	ASSERTV(pwc!=NULL);
	ASSERTV(pData!=NULL);
	ASSERTV(rFull!=NULL);
	ASSERTV(rTitle!=NULL);

	DI;

	/* Handle switch between title/back here because it only has to */ 
	/* happen when titlebars are used, and the time taken is the    */ 
	/* same no matter where it happens.  This way, with no titlebar */ 
	/* it doesn't slow things down.                                 */ 
	/* (also these aren't initialized without bTitlebars)           */ 
	SelectObject(dc,pwc->brTitle);
	SelectObject(dc,pwc->pTitleBorder);

	if (!GlobalData.bBevel) {
		/* with no bevel, just draw over the whole rTitle */ 
		CI Rectangle(dc,rTitle->left,rTitle->top,rTitle->right,rTitle->bottom-1);
		CTRACE(TRACE_WARNING,!i,"PWT_C: Rectangle failed.");
	} else {
		/* with bevel, don't draw on the left, top, or right edges */ 
		/* but do draw on the bottom edge                          */ 
		/* this matches with the area bevelled by PaintWnd_Colors  */ 
		CI Rectangle(dc,rTitle->left+1,rTitle->top+1,rTitle->right-1,rTitle->bottom-1);
		CTRACE(TRACE_WARNING,!i,"PWT_C: Rectangle failed.");
		CI0;

		SelectObject(dc,pwc->pTitleLight);
		CI MoveToEx(dc,rTitle->left+1,rTitle->bottom-2,NULL);
		CI LineTo(dc,rTitle->left+1,rTitle->top+1); /* left edge */ 
		CI LineTo(dc,rTitle->right-2,rTitle->top+1); /* top edge */ 
		SelectObject(dc,pwc->pTitleDark);
		CI LineTo(dc,rTitle->right-2,rTitle->bottom-2); /* right edge */ 
		CI LineTo(dc,rTitle->left+1,rTitle->bottom-2); /* bottom edge */ 

		CTRACE(TRACE_WARNING,!i,"PWT_C: Failed to draw bevel.");
	}

	/* switch back to main miniwindow pen/brush */ 
	SelectObject(dc,pwc->brBack);
	SelectObject(dc,pwc->pBackBorder);
}

void EndPaintWnd_Colors(HDC dc,PixelBlock &pb,void *store)
{
	s_PaintWnd_Colors *pwc = (s_PaintWnd_Colors *)store;

	ASSERTV(store!=NULL);
	ASSERTV(pwc!=NULL);

	DI;

	/* select the stuff out of the device context */ 
	SelectObject(dc,pwc->oldBrush);
	SelectObject(dc,pwc->oldPen);

	/* delete them */ 
	CI DeleteObject(pwc->brBack); StatGDIDel(Brush);
	CI DeleteObject(pwc->pBackBorder); StatGDIDel(Pen);
	CI DeleteObject(pwc->pBackDark); StatGDIDel(Pen);
	CI DeleteObject(pwc->pBackLight); StatGDIDel(Pen);
	if (GlobalData.bTitlebars) {
		CI DeleteObject(pwc->brTitle); StatGDIDel(Brush);
		CI DeleteObject(pwc->pTitleBorder); StatGDIDel(Pen);
		CI DeleteObject(pwc->pTitleDark); StatGDIDel(Pen);
		CI DeleteObject(pwc->pTitleLight); StatGDIDel(Pen);
	}
	CTRACE(TRACE_WARNING,!i,"EPW_C: Failed to delete handles.");
}
