/*
  RabidVWM LiteStep Module - version 1.0
	WndProc

  Copyright (C) 2002 Joshua Seagoe
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include "RabidVWM.h"
#include <Windowsx.h>
#include <Shellapi.h>
#include <tchar.h>

// update the hover status on mouse movement
// the other half of hover detection is in WM_TIMER
static void UpdateHover(bool bSetTimer)
{
	POINT pt;
	/* set mouse over */ 
	GetCursorPos(&pt);
	if (GlobalData.HoverPoint.x != pt.x
		||GlobalData.HoverPoint.y != pt.y) {
		if (!GlobalData.bMouseOver) {
			GlobalData.bMouseOver = true;
			if (GlobalData.OnMouseOver) {
				AtVarExec(GlobalData.OnMouseOver,true,true);
			}
		}
		/* clear hover flag            */ 
		if (GlobalData.bMouseHovering) {
			GlobalData.bMouseHovering = false;
			if (GlobalData.OnUnhover && !GlobalData.bInDragDrop) {
				GlobalData.OnUnhover->MousePos = GlobalData.HoverPoint;
				ScreenToClient(GlobalData.hMainWnd,&GlobalData.OnUnhover->MousePos);
				AtVarExec(GlobalData.OnUnhover,false,true);
			}
		}
		if (!bSetTimer) {
			KillTimer(GlobalData.hMainWnd,GlobalData.HoverTimerID);
		} else {
			/* timer for mouse out/hover */ 
			GlobalData.HoverTimerID = SetTimer(
				GlobalData.hMainWnd,TIMER_HOVER,GlobalData.HoverTime,NULL);
			GlobalData.HoverPoint = pt;
			CTRACE(TRACE_WARNING,!GlobalData.HoverTimerID,"Unable to set HoverTimer.");
		}
	}
}

static void CancelMouse(int which,int x,int y)
{
	ASSERTV(which<=MOUSE_ACTION_COUNT);
	ASSERTV(GlobalData.MouseCurrent<=MOUSE_ACTION_COUNT);
	/* cancel any action in progress */ 
	if (GlobalData.MouseCurrent!=0 &&
		MouseTable[GlobalData.MouseAction[GlobalData.MouseCurrent-1]].Cancel!=NULL)
		MouseTable[GlobalData.MouseAction[GlobalData.MouseCurrent-1]].Cancel(x,y);
	/* no action in progress         */ 
	GlobalData.MouseCurrent = which;
}

// route mouse down to proper handler
// which is ** 1 based **
static void DoMouseDown(int which,int x,int y)
{
	ASSERTV(which<=MOUSE_ACTION_COUNT);
	ASSERTV(GlobalData.MouseCurrent<=MOUSE_ACTION_COUNT);
	CancelMouse(which,x,y);
	/* call the action */ 
	if (MouseTable[GlobalData.MouseAction[which-1]].Press!=NULL)
		MouseTable[GlobalData.MouseAction[which-1]].Press(x,y);

	UpdateHover(false);
}

// route mouse up to proper handler
// which is ** 1 based **
static void DoMouseUp(int which,int x,int y)
{
	ASSERTV(which<=MOUSE_ACTION_COUNT);
	ASSERTV(GlobalData.MouseCurrent<=MOUSE_ACTION_COUNT);
	ASSERTV(GlobalData.MouseCurrent==0 || GlobalData.MouseCurrent==which);
	(void)which; /* use current instead */ 

	if (GlobalData.MouseCurrent!=0 &&
		MouseTable[GlobalData.MouseAction[GlobalData.MouseCurrent-1]].Release!=NULL)
		MouseTable[GlobalData.MouseAction[GlobalData.MouseCurrent-1]].Release(x,y);
	GlobalData.MouseCurrent = 0;

	UpdateHover(true);
}

// the big bad boy himself
LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	const char * const easterstring = "MOO";
	static const char *easterptr = easterstring;

#ifdef VWM_STATS
	int WndProcStart = GetTickCount();
	if (VWMStats.WndProcLevel++ == 0)
		VWMStats.WndProcStart = WndProcStart;
#endif
	switch (message) {

	case WM_CREATE:
		TRACE(TRACE_STATUS,"WM_CREATE");
		/* save main window handle */ 
		GlobalData.hMainWnd = hwnd;
		/* allocate back buffer    */ 
		Resize(GlobalData.VWMSize.cx,GlobalData.VWMSize.cy);
#ifdef VWM_STATS
		VWMStats.CreateTime = GetTickCount()-WndProcStart;
#endif
		break;

	case WM_CLOSE:
		TRACE(TRACE_STATUS,"WM_CLOSE");
		/* don't close on alt-f4 */ 
		break;

	case WM_CHAR:
	case WM_KEYDOWN: 
		TRACE(TRACE_STATUS,"WM_KEYDOWN");
		/* check if it's escape during a drag and cancel */ 
		if (wParam==VK_ESCAPE && GlobalData.MouseCurrent!=0) {
			CancelMouse(0,0,0);
			break;
		}
    /* should we do the easter egg? */ 
		if (wParam>='A' && wParam<='Z')
		{
			if (wParam==(unsigned)*easterptr){if
      (*++easterptr==0)StartEaster();}else
      {StopEaster();easterptr=easterstring

      ;}

		}
#if !defined(NDEBUG) && defined(TRACE_FILE)
		switch (wParam) {
		/* debug stuff */ 
		case '8':
			DumpStarLists();
			break;
		case 'A':
			DumpActions();
			break;
		case 'B':
			DumpBools();
			break;
#ifdef VWM_STATS
		case 'S':
			DumpStats();
			break;
#endif
		case 'W':
			DumpWindowList();
			break;
		}
		break;
#endif
	case WM_KEYUP:
		/* we don't respond directly to any keys, forward them on to */ 
		/* LiteStep (the desktop window, our parent) so they'll work */ 
		PostMessage(GlobalData.hParentWnd,message,wParam,lParam);
		break;

// ************ BEGIN MOUSE EVENTS
		/* mouse button actions are selected by table    */ 
	case WM_LBUTTONDOWN: 
		AtVarExec(GlobalData.OnClickLeft,true,true);
		DoMouseDown(1,GET_X_LPARAM(lParam),GET_Y_LPARAM(lParam)); break;
	case WM_LBUTTONUP:   DoMouseUp  (1,GET_X_LPARAM(lParam),GET_Y_LPARAM(lParam)); break;
	case WM_RBUTTONDOWN:
		AtVarExec(GlobalData.OnClickRight,true,true);
		DoMouseDown(2,GET_X_LPARAM(lParam),GET_Y_LPARAM(lParam)); break;
	case WM_RBUTTONUP:   DoMouseUp  (2,GET_X_LPARAM(lParam),GET_Y_LPARAM(lParam)); break;
	case WM_MBUTTONDOWN: 
		AtVarExec(GlobalData.OnClickMiddle,true,true);
		DoMouseDown(3,GET_X_LPARAM(lParam),GET_Y_LPARAM(lParam)); break;
	case WM_MBUTTONUP:   DoMouseUp  (3,GET_X_LPARAM(lParam),GET_Y_LPARAM(lParam)); break;
	case WM_MOUSEMOVE:
		/* route to the mouse action in progress, if any */ 
		if (GlobalData.MouseCurrent!=0 && 
			MouseTable[GlobalData.MouseAction[GlobalData.MouseCurrent-1]].Move!=NULL)
    {
			MouseTable[GlobalData.MouseAction[GlobalData.MouseCurrent-1]].
				Move(GET_X_LPARAM(lParam),GET_Y_LPARAM(lParam));
		}
		UpdateHover(true);
    AtVarExec(GlobalData.OnMouseMove,true,true);
		break;

	case WM_CAPTURECHANGED:
		TRACE(TRACE_STATUS,"WM_CAPTURECHANGED");
		CancelMouse(0,0,0);
		break;

#ifdef INCLUDE_SCROLL_EVENTS
	case WM_MOUSEWHEEL:
		{
			static int ScrollDist = 0;
			ScrollDist += GET_WHEEL_DELTA_WPARAM(wParam);
			int notches = ScrollDist/WHEEL_DELTA;

			if (notches>0) {
				if (GlobalData.OnScrollUp) {
					GlobalData.AtParams[0] = notches;
					GlobalData.OnScrollUp->MousePos.x = GET_X_LPARAM(lParam);
					GlobalData.OnScrollUp->MousePos.y = GET_Y_LPARAM(lParam);
					AtVarExec(GlobalData.OnScrollUp,false,true);
				}
			} else if (notches<0) {
				if (GlobalData.OnScrollDown) {
					GlobalData.AtParams[0] = -notches;
					GlobalData.OnScrollDown->MousePos.x = GET_X_LPARAM(lParam);
					GlobalData.OnScrollDown->MousePos.y = GET_Y_LPARAM(lParam);
					AtVarExec(GlobalData.OnScrollDown,false,true);
				}
			}
			ScrollDist -= WHEEL_DELTA*notches;
		}
		break;
#endif
// ************ END MOUSE EVENTS

	case WM_ERASEBKGND:
		/* erasing the background creates flicker                */ 
		/* we're bufferred anyway, so it doesn't help any either */ 
		break;

	case WM_PAINT:
		ASSERTV(!GlobalData.bTransparent);
		{
			/* ready  */ 
			PAINTSTRUCT ps;
			HDC hdc = BeginPaint(GlobalData.hMainWnd,&ps);
			ASSERTV(hdc!=NULL);
			HDC mdc = CreateCompatibleDC(hdc);
			ASSERTV(mdc!=NULL);
			/* aim    */ 
			HBITMAP oldbm = (HBITMAP)SelectObject(mdc,GlobalData.bmBuffer);
			DI;

			/* fire   */ 
			CI BitBlt(hdc,0,0,GlobalData.VWMSize.cx,GlobalData.VWMSize.cy,mdc,0,0,SRCCOPY);

			CTRACE(TRACE_WARNING,!i,"Can't blit to main window.");

			/* reload */ 
			SelectObject(mdc,oldbm);
			DeleteDC(mdc);
			EndPaint(GlobalData.hMainWnd,&ps);
		}
		break;

	case WM_TIMER:
		if (wParam==TIMER_REFRESH) {
			UpdateWindowList(); /* make sure windows are up to date */ 
			if (GlobalData.bItsEaster) {
				StepEaster();    /* forces a repaint on every refresh */ 
			} else
			if (GlobalData.bChangeState || GlobalData.bConstantUpdate) {
				Repaint();        /* use common code to redraw        */ 
			}
			/* this makes flashing things flash           */ 
			GlobalData.bFlashState = !GlobalData.bFlashState;

#ifdef VWM_STATS
			VWMStats.RefreshTime += GetTickCount()-WndProcStart;
#endif
		} else if (wParam==TIMER_HOVER) {
			POINT pt;
			/* the mouse is either hovering or has left   */ 
			/* get the cursor position to determine which */ 
			GetCursorPos(&pt);
			if (pt.x == GlobalData.HoverPoint.x
				&&pt.y == GlobalData.HoverPoint.y) {
				/* it's hovering */ 
				if (!GlobalData.bMouseHovering) {
					GlobalData.bMouseHovering = true;
					if (GlobalData.bInDragDrop) {
						/* switch desks if you hover while dragging */ 
						TRACE(TRACE_STATUS,"Drag&Drop Hover");
						ScreenToClient(GlobalData.hMainWnd,&pt);
						SwitchDesk(DeskFromVWMPoint(pt.x,pt.y));
					} else if (GlobalData.OnHover) {
						AtVarExec(GlobalData.OnHover,true,true);
					}
				}
			} else {
				/* it's left                             */ 
				/* if it was hovering, it sure isn't now */ 
				if (GlobalData.bMouseHovering) {
					GlobalData.bMouseHovering = false;
					if (GlobalData.OnUnhover && !GlobalData.bInDragDrop) {
						GlobalData.OnUnhover->MousePos = GlobalData.HoverPoint;
						ScreenToClient(GlobalData.hMainWnd,&GlobalData.OnUnhover->MousePos);
						AtVarExec(GlobalData.OnUnhover,false,true);
					}
				}
				GlobalData.bMouseOver = false;
				if (GlobalData.OnMouseOut) {
					GlobalData.OnMouseOut->MousePos = GlobalData.HoverPoint;
					ScreenToClient(GlobalData.hMainWnd,&GlobalData.OnMouseOut->MousePos);
					AtVarExec(GlobalData.OnMouseOut,false,true);
				}
			}
			KillTimer(GlobalData.hMainWnd,GlobalData.HoverTimerID);
		} else if (wParam==TIMER_FOCUS) {
			/* last time we tried setting focus it failed, try again     */ 
			SetDeskFocus();
		} else if (wParam==TIMER_SNAPSHOT) {
			/* take a snapshot for slow-painting windows                 */ 
			TRACE(TRACE_STATUS,"TIMER_SNAPSHOT");
			GlobalData.bDelaySnapshot = false;
			Repaint();
			KillTimer(GlobalData.hMainWnd,GlobalData.SnapShotTimerID);
		}
		break;

#ifndef VWMDROPTARGET
	case WM_DROPFILES:
		TRACE(TRACE_STATUS,"WM_DROPFILES");
		{ /* another thing stolen from sysvwm                          */ 
			/* changed it juuust enough that they don't get royalties ;) */ 
			int nDropped,i;
			char szFname[256];
			POINT pt;

			/* get desk           */ 
			if (!DragQueryPoint((HDROP) wParam, &pt))
				GetCursorPos(&pt);
			SwitchDesk(DeskFromVWMPoint(pt.x,pt.y));

			/* open dropped files */ 
			nDropped = DragQueryFile((HDROP) wParam, 0xFFFFFFFF, NULL, 0);
			for (i = 0; i < nDropped; i++)
			{
				DragQueryFile((HDROP) wParam, i, (char *)&szFname, sizeof(szFname));
				if (szFname && szFname[0])
				{
          LSExecuteEx(NULL, NULL, szFname, NULL, NULL, SW_SHOWNORMAL);
				}
			}
		}
		break;
#endif

	case WM_DISPLAYCHANGE:
	case WM_SETTINGCHANGE:
		UpdateDesktopBits();
		Repaint();
		/* Another thought is to do a full refresh to get */ 
		/* the multimonitor settings (and single monitor  */ 
		/* resolution) again, but that would clear all    */ 
		/* the "on the fly" changes with no warning.      */ 
		/* So let them call !VWMReset if they want that.  */ 
		break;

	case LM_GETREVID: 
		TRACE(TRACE_STATUS,"LM_GETREVID");
		{
			PTSTR buffer = (PTSTR)lParam;
			
			if (wParam == 0)
			{
				_tcscpy(buffer, szAppName);
				_tcscat(buffer, _T(".dll: "));
				_tcscat(buffer, szAppVer);
			}
			else
			{
				_tcscpy(buffer, _T(""));
			}
			TRACE(TRACE_STATUS,buffer);
#ifdef VWM_STATS
		if (--VWMStats.WndProcLevel == 0)
			VWMStats.WndProcTime += GetTickCount() - VWMStats.WndProcStart;
#endif
			return _tcslen(buffer);
		}

		/* we could ignore these and rely on the timer-based */ 
		/* refresh, but this will supply a faster response   */ 
		/* when windows are created/destroyed/change z-order */ 
	case LM_WINDOWDESTROYED:
    /* do NOT just remove the window from the wnd list.  */ 
    /* this would be faster, of course, but would make   */ 
    /* the pointers for the paint loop invalid.          */ 
#if !defined(NDEBUG) && defined(TRACE_FILE)
    {
      s_WinData *pData = GlobalData.KnownWindows + FindKnownWindow((HWND)wParam);
      if (pData->hWnd!=(HWND)wParam) pData = NULL;
		  WTRACE(TRACE_STATUS,pData,"LM_WINDOWDESTROYED");
    }
#endif
		UpdateWindowList(); /* make sure windows are up to date */ 
    {
      s_WinData *pData = GlobalData.KnownWindows + FindKnownWindow((HWND)wParam);
      /* if the window was really destroyed... */ 
      if (pData->hWnd!=(HWND)wParam)
		    SetDeskFocus();     /* try to keep focus on this desk   */ 
    }
	case LM_WINDOWACTIVATED:
#if !defined(NDEBUG) && defined(TRACE_FILE)
    if (message==LM_WINDOWACTIVATED)
    {
      s_WinData *pData = GlobalData.KnownWindows + FindKnownWindow((HWND)wParam);
      if (pData->hWnd!=(HWND)wParam) pData = NULL;
		  WTRACE(TRACE_STATUS,pData,"LM_WINDOWACTIVATED");
    }
#endif
	case LM_WINDOWCREATED:
#if !defined(NDEBUG) && defined(TRACE_FILE)
    if (message==LM_WINDOWCREATED)
    {
      s_WinData *pData = GlobalData.KnownWindows + FindKnownWindow((HWND)wParam);
      if (pData->hWnd!=(HWND)wParam) pData = NULL;
		  WTRACE(TRACE_STATUS,pData,"LM_WINDOWCREATED");
    }
#endif
		UpdateWindowList(); /* make sure windows are up to date */ 
		if (GlobalData.bChangeState)
		{										/* delay window snapshots           */ 
			GlobalData.bDelaySnapshot = true;
			Repaint();        /* use common code to redraw        */ 
		}
		break;

	case LM_BRINGTOFRONT:
		TRACE(TRACE_STATUS,"LM_BRINGTOFRONT");
		if (!GlobalData.bNoSwitchOnFocus)
		{
			s_WinData *pData;
			/* find the window  */ 
			pData = &GlobalData.KnownWindows[FindKnownWindow((HWND)lParam)];
			if (pData->hWnd == (HWND)lParam) {
				if (pData->HomeDesk<0) {
					if (IsWndMinimized(pData)) {
						/* minimized, restore it              */ 
						ShowWindow(pData->hWnd, SW_SHOWDEFAULT);
						SetHomeDesk(pData,GlobalData.CurrentDesk);
					} else {
						TRACE(TRACE_STATUS,"Switching to lost window.");
						/* if it's lost, move the window here */ 
						MoveWindowToDesk(pData, GlobalData.CurrentDesk, NULL);
					}
				} else {
					if (pData->HomeDesk != GlobalData.CurrentDesk)
						/* switch to its homedesk             */ 
						SwitchDesk(pData->HomeDesk);
				}
			}
			/* give it focus    */ 
			/* switch for winnt */ 
			if (SwitchToThisWindow) SwitchToThisWindow((HWND)lParam,TRUE);
			/* switch for win9x */ 
			SetForegroundWindow((HWND)lParam);
		}
		break;

	case LM_SWITCHTON:
		TRACE(TRACE_STATUS,"LM_SWITCHTON");
		/* SwitchDesk will check for validity */ 
		SwitchDesk(wParam);
		break;

	case LM_REDRAW:
		{
			s_WinData *pData = GetWindowPtr((HWND)wParam);
			if (GlobalData.OnFlash && GlobalData.OnFlash->Command
				/* only trigger when flash is turned on          */ 
				&& lParam==TRUE && pData!=NULL
				&& !(pData->bFlash || pData->bFlashOn))
			{
				GlobalData.OnFlash->pData = pData;
				AtVarExec(GlobalData.OnFlash,true,false);
			}
			if (!GlobalData.bNoShellFlash) {
				/* wParam = hWnd, lParam = bFlash                */ 
				/* method stolen from taskbar.dll (thanks Chaku) */ 

				/* Don't clear flash from here, my concept of flashing is different.   */ 
				/* The shell calls this with set & clear alternating, where I just     */ 
				/* (with the newer HSHELL_FLASH, it calls it with "toggle"...)         */ 
				/* want it set, and flash on its own.  This means windows can't clear  */ 
				/* their flash state, but if that's a problem I'll deal with it later. */ 
				if (lParam==TRUE && pData->hWnd==(HWND)wParam
					&& !(pData->bFlash || pData->bFlashOn))
				{
					pData->bFlash = true;
					pData->bFlashOn = false;
					pData->iFlashCountLeft = GlobalData.iFlashCount;
				}
				/* Don't repaint here, wait for timer.           */ 
				/* Otherwise you get annoying flicker            */ 

				/* note that I'm checking lParam against TRUE rather than false to     */ 
				/* avoid going flash-crazy with a LiteStep core later than 11-19-01    */ 
				/* that doesn't also process the HSHELL_FLASH message.                 */ 
				/* (it sends the handle again as lParam.)                              */ 
				/* unfortunately, this means we get no flashes, and sometimes RabidVWM */ 
				/* seems to be at fault because other modules "work"                   */ 
			}
		}
		break;

	case LM_REFRESH:
		TRACE(TRACE_STATUS,"LM_REFRESH");
		Refresh();
		break;

	case WM_DESTROY:
		TRACE(TRACE_STATUS,"WM_DESTROY");
		if (GlobalData.bLoadInBox/* && GlobalData.RefreshTimerID*/) LocalQuit();
		break;
		
	default:
#ifdef VWM_STATS
		--VWMStats.WndProcLevel;
#endif
		return DefWindowProc(hwnd,message,wParam,lParam);
	}

#ifdef VWM_STATS
	if (--VWMStats.WndProcLevel == 0)
		VWMStats.WndProcTime += GetTickCount() - VWMStats.WndProcStart;
#endif
	return 0;
}

